from abc import abstractmethod

from .abstract_problem import AbstractProblem


class TimeDependentProblem(AbstractProblem):

    @property
    @abstractmethod
    def temporal_variable(self):
        pass



import argparse
import torch
from torch.nn import Softplus

from pina import PINN, Plotter
from pina.model import FeedForward
from problems.burgers import Burgers1D


class myFeature(torch.nn.Module):
    """
    Feature: sin(pi*x)
    """
    def __init__(self, idx):
        super(myFeature, self).__init__()
        self.idx = idx

    def forward(self, x):
        return torch.sin(torch.pi * x[:, self.idx])


if __name__ == "__main__":

    parser = argparse.ArgumentParser(description="Run PINA")
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument("-s", "-save", action="store_true")
    group.add_argument("-l", "-load", action="store_true")
    parser.add_argument("id_run", help="number of run", type=int)
    parser.add_argument("features", help="extra features", type=int)
    args = parser.parse_args()

    feat = [myFeature(0)] if args.features else []

    burgers_problem = Burgers1D()
    model = FeedForward(
        layers=[30, 20, 10, 5],
        output_variables=burgers_problem.output_variables,
        input_variables=burgers_problem.input_variables,
        func=Softplus,
        extra_features=feat,
    )

    pinn = PINN(
        burgers_problem,
        model,
        lr=0.006,
        error_norm='mse',
        regularizer=0,
        lr_accelerate=None)

    if args.s:
        pinn.span_pts(2000, 'latin', ['D'])
        pinn.span_pts(150, 'random', ['gamma1', 'gamma2', 't0'])
        pinn.train(5000, 100)
        pinn.save_state('pina.burger.{}.{}'.format(args.id_run, args.features))
    else:
        pinn.load_state('pina.burger.{}.{}'.format(args.id_run, args.features))
        plotter = Plotter()
        plotter.plot(pinn)



import torch

from pina.problem import TimeDependentProblem, SpatialProblem
from pina.operators import grad
from pina import Condition
from pina.span import Span


class Burgers1D(TimeDependentProblem, SpatialProblem):

    spatial_variables = ['x']
    temporal_variable = ['t']
    output_variables = ['u']
    domain = Span({'x': [-1, 1], 't': [0, 1]})

    def burger_equation(input_, output_):
        grad_u = grad(output_['u'], input_)
        grad_x = grad_u['x']
        grad_t = grad_u['t']
        gradgrad_u_x = grad(grad_u['x'], input_)
        return (
            grad_u['t'] + output_['u']*grad_u['x'] -
            (0.01/torch.pi)*gradgrad_u_x['x']
        )

    def nil_dirichlet(input_, output_):
        u_expected = 0.0
        return output_['u'] - u_expected

    def initial_condition(input_, output_):
        u_expected = -torch.sin(torch.pi*input_['x'])
        return output_['u'] - u_expected

    conditions = {
        'gamma1': Condition(Span({'x': -1, 't': [0, 1]}), nil_dirichlet),
        'gamma2': Condition(Span({'x':  1, 't': [0, 1]}), nil_dirichlet),
        't0': Condition(Span({'x': [-1, 1], 't': 0}), initial_condition),
        'D': Condition(Span({'x': [-1, 1], 't': [0, 1]}), burger_equation),
    }



import torch
import numpy as np

from .problem import AbstractProblem
from . import PINN, LabelTensor
torch.pi = torch.acos(torch.zeros(1)).item() * 2  # 3.1415927410125732


class ParametricPINN(PINN):

    def __init__(self, problem, model, optimizer=torch.optim.Adam, lr=0.001,
                 regularizer=0.00001, data_weight=1., dtype=torch.float64,
                 device='cpu', lr_accelerate=None, error_norm='mse'):
        '''
        :param Problem problem: the formualation of the problem.
        :param dict architecture: a dictionary containing the information to
            build the model. Valid options are:
            - inner_size [int] the number of neurons in the hidden layers; by
              default is 20.
            - n_layers [int] the number of hidden layers; by default is 4.
            - func [nn.Module or str] the activation function; passing a `str`
              is possible to chose adaptive function (between 'adapt_tanh'); by
              default is non-adaptive iperbolic tangent.
        :param float lr: the learning rate; default is 0.001
        :param float regularizer: the coefficient for L2 regularizer term
        :param type dtype: the data type to use for the model. Valid option are
            `torch.float32` and `torch.float64` (`torch.float16` only on GPU);
            default is `torch.float64`.
        :param float lr_accelete: the coefficient that controls the learning
            rate increase, such that, for all the epoches in which the loss is
            decreasing, the learning_rate is update using
                $learning_rate = learning_rate * lr_accelerate$.
            When the loss stops to decrease, the learning rate is set to the
            initial value [TODO test parameters]

        '''

        self.problem = problem


        # self._architecture = architecture if architecture else dict()
        # self._architecture['input_dimension'] = self.problem.domain_bound.shape[0]
        # self._architecture['output_dimension'] = len(self.problem.variables)
        # if hasattr(self.problem, 'params_domain'):
            # self._architecture['input_dimension'] += self.problem.params_domain.shape[0]

        self.accelerate = lr_accelerate

        self.error_norm = error_norm

        if device == 'cuda' and not torch.cuda.is_available():
            raise RuntimeError
        self.device = torch.device(device)

        self.dtype = dtype
        self.history = []

        self.model = model
        self.model.to(dtype=self.dtype, device=self.device)

        self.input_pts = {}
        self.truth_values = {}


        self.trained_epoch = 0
        self.optimizer = optimizer(
            self.model.parameters(), lr=lr, weight_decay=regularizer)

        self.data_weight = data_weight

    @property
    def problem(self):
        return self._problem

    @problem.setter
    def problem(self, problem):
        if not isinstance(problem, AbstractProblem):
            raise TypeError
        self._problem = problem

    def get_data_residuals(self):

        data_residuals = []

        for output in self.data_pts:
            data_values_pred = self.model(self.data_pts[output])
            data_residuals.append(data_values_pred - self.data_values[output])

        return torch.cat(data_residuals)

    def get_phys_residuals(self):
        """
        """

        residuals = []
        for equation in self.problem.equation:
            residuals.append(equation(self.phys_pts, self.model(self.phys_pts)))
        return residuals



    def train(self, stop=100, frequency_print=2, trial=None):

        epoch = 0

        while True:

            losses = []

            for condition_name in self.problem.conditions:
                condition = self.problem.conditions[condition_name]
                pts = self.input_pts[condition_name]

                predicted = self.model(pts.tensor)
                #predicted = self.model(pts)

                residuals = condition.function(pts, predicted)
                losses.append(self._compute_norm(residuals))
            self.optimizer.zero_grad()
            sum(losses).backward()
            self.optimizer.step()

                #for p in parameters:
                #    pts = self.input_pts[condition_name]
                #    #pts = torch.cat([pts.tensor, p.double().repeat(pts.tensor.shape[0]).reshape(-1, 2)], axis=1)
                #    #pts = torch.cat([pts.tensor, p.double().repeat(pts.tensor.shape[0]).reshape(-1, 1)], axis=1)
                #    #print(self.problem.input_variables)
                #    # print(self.problem.parameters)
                #    # print(pts.shape)
                #    print(pts.tensor.repeat_interleave(parameters.shape[0]))
                #    # print(pts)
                #    # gg
                #    a = torch.cat([
                #        pts.tensor.repeat_interleave(parameters.shape[0], dim=0),
                #        torch.tile(parameters, (pts.tensor.shape[0], 1))
                #    ], axis=1)
                #    for i in a:
                #        print(i.detach())
                #    ttt
                #    pts = LabelTensor(pts, self.problem.input_variables + self.problem.parameters)


                #    ffff
                #    print(pts.labels)

                #    predicted = self.model(pts.tensor)
                #    #predicted = self.model(pts)
                #    if isinstance(self.problem.conditions[condition_name]['func'], list):
                #        for func in self.problem.conditions[condition_name]['func']:
                #            residuals = func(pts, LabelTensor(p.reshape(1, -1), ['mu', 'alpha']), predicted)
                #            tmp_losses.append(self._compute_norm(residuals))
                #    else:
                #        residuals = self.problem.conditions[condition_name]['func'](pts, p, predicted)
                #        tmp_losses.append(self._compute_norm(residuals))
                #losses.append(sum(tmp_losses))


            self.trained_epoch += 1
            #if epoch % 10 == 0:
            #    self.history.append(losses)

            epoch += 1

            if trial:
                import optuna
                trial.report(loss[0].item()+loss[1].item(), epoch)
                if trial.should_prune():
                    raise optuna.exceptions.TrialPruned()

            if isinstance(stop, int):
                if epoch == stop:
                    break
            elif isinstance(stop, float):
                if loss[0].item() + loss[1].item() < stop:
                    break

            if epoch % frequency_print == 0:
                print('[epoch {:05d}] {:.6e} '.format(self.trained_epoch, sum(losses).item()), end='')
                for loss in losses:
                    print('{:.6e} '.format(loss), end='')
                print()

        return sum(losses).item()


    def error(self, dtype='l2', res=100):

        import numpy as np
        if hasattr(self.problem, 'truth_solution') and self.problem.truth_solution is not None:
            pts_container = []
            for mn, mx in self.problem.domain_bound:
                pts_container.append(np.linspace(mn, mx, res))
            grids_container = np.meshgrid(*pts_container)
            Z_true = self.problem.truth_solution(*grids_container)

        elif hasattr(self.problem, 'data_solution') and self.problem.data_solution is not None:
            grids_container = self.problem.data_solution['grid']
            Z_true = self.problem.data_solution['grid_solution']
        try:
            unrolled_pts = torch.tensor([t.flatten() for t in grids_container]).T.to(dtype=self.dtype, device=self.device)
            Z_pred = self.model(unrolled_pts)
            Z_pred = Z_pred.detach().numpy().reshape(grids_container[0].shape)

            if dtype == 'l2':
                return np.linalg.norm(Z_pred - Z_true)/np.linalg.norm(Z_true)
            else:
                # TODO H1
                pass
        except:
            print("")
            print("Something went wrong...")
            print("Not able to compute the error. Please pass a data solution or a true solution")

    def plot(self, res, param, filename=None, variable=None):
            '''
            '''
            import matplotlib
            matplotlib.use('GTK3Agg')
            import matplotlib.pyplot as plt

            pts_container = []
            for mn, mx in [[-1, 1], [-1, 1]]:
                pts_container.append(np.linspace(mn, mx, res))
            grids_container = np.meshgrid(*pts_container)
            unrolled_pts = torch.tensor([t.flatten() for t in grids_container]).T
            unrolled_pts = torch.cat([unrolled_pts, param.double().repeat(unrolled_pts.shape[0]).reshape(-1, 2)], axis=1)

            unrolled_pts.to(dtype=self.dtype)
            unrolled_pts = LabelTensor(unrolled_pts, ['x1', 'x2', 'mu1', 'mu2'])

            Z_pred = self.model(unrolled_pts.tensor)
            n = Z_pred.tensor.shape[1]
            plt.figure(figsize=(6*n, 6))

            for i, output in enumerate(Z_pred.tensor.T, start=1):
                output = output.detach().numpy().reshape(grids_container[0].shape)
                plt.subplot(1, n, i)
                plt.contourf(*grids_container, output)
                plt.colorbar()

            if filename is None:
                plt.show()
            else:
                plt.savefig(filename)

    def plot_params(self, res, param, filename=None, variable=None):
            '''
            '''
            import matplotlib
            matplotlib.use('GTK3Agg')
            import matplotlib.pyplot as plt

            if hasattr(self.problem, 'truth_solution') and self.problem.truth_solution is not None:
                n_plot = 2
            elif hasattr(self.problem, 'data_solution') and self.problem.data_solution is not None:
                n_plot = 2
            else:
                n_plot = 1

            fig, axs = plt.subplots(nrows=1, ncols=n_plot, figsize=(n_plot*6,4))
            if not isinstance(axs, np.ndarray): axs = [axs]

            if hasattr(self.problem, 'data_solution') and self.problem.data_solution is not None:
                grids_container = self.problem.data_solution['grid']
                Z_true = self.problem.data_solution['grid_solution']
            elif hasattr(self.problem, 'truth_solution') and self.problem.truth_solution is not None:

                pts_container = []
                for mn, mx in self.problem.domain_bound:
                    pts_container.append(np.linspace(mn, mx, res))

                grids_container = np.meshgrid(*pts_container)
                Z_true = self.problem.truth_solution(*grids_container)

            pts_container = []
            for mn, mx in self.problem.domain_bound:
                pts_container.append(np.linspace(mn, mx, res))
            grids_container = np.meshgrid(*pts_container)
            unrolled_pts = torch.tensor([t.flatten() for t in grids_container]).T.to(dtype=self.type)
            #print(unrolled_pts)
            #print(param)
            param_unrolled_pts = torch.cat((unrolled_pts, param.repeat(unrolled_pts.shape[0], 1)), 1)
            if variable==None:
                variable = self.problem.variables[0]
                Z_pred = self.evaluate(param_unrolled_pts)[variable]
                variable = "Solution"
            else:
                Z_pred = self.evaluate(param_unrolled_pts)[variable]

            Z_pred= Z_pred.detach().numpy().reshape(grids_container[0].shape)
            set_pred = axs[0].contourf(*grids_container, Z_pred)
            axs[0].set_title('PINN [trained epoch = {}]'.format(self.trained_epoch) + " " + variable) #TODO add info about parameter in the title
            fig.colorbar(set_pred, ax=axs[0])

            if n_plot == 2:

                set_true = axs[1].contourf(*grids_container, Z_true)

                axs[1].set_title('Truth solution')
                fig.colorbar(set_true, ax=axs[1])

            if filename is None:
                    plt.show()
            else:
                    fig.savefig(filename + " " + variable)

    def plot_error(self, res, filename=None):
        import matplotlib
        matplotlib.use('GTK3Agg')
        import matplotlib.pyplot as plt


        fig, axs = plt.subplots(nrows=1, ncols=1, figsize=(6,4))
        if not isinstance(axs, np.ndarray): axs = [axs]

        if hasattr(self.problem, 'data_solution') and self.problem.data_solution is not None:
            grids_container = self.problem.data_solution['grid']
            Z_true = self.problem.data_solution['grid_solution']
        elif hasattr(self.problem, 'truth_solution') and self.problem.truth_solution is not None:
            pts_container = []
            for mn, mx in self.problem.domain_bound:
                pts_container.append(np.linspace(mn, mx, res))

            grids_container = np.meshgrid(*pts_container)
            Z_true = self.problem.truth_solution(*grids_container)
        try:
            unrolled_pts = torch.tensor([t.flatten() for t in grids_container]).T.to(dtype=self.type)

            Z_pred = self.model(unrolled_pts)
            Z_pred = Z_pred.detach().numpy().reshape(grids_container[0].shape)
            set_pred = axs[0].contourf(*grids_container, abs(Z_pred - Z_true))
            axs[0].set_title('PINN [trained epoch = {}]'.format(self.trained_epoch) + "Pointwise Error")
            fig.colorbar(set_pred, ax=axs[0])

            if filename is None:
                    plt.show()
            else:
                    fig.savefig(filename)
        except:
            print("")
            print("Something went wrong...")
            print("Not able to plot the error. Please pass a data solution or a true solution")

'''
print(self.pred_loss.item(),loss.item(), self.old_loss.item())
if self.accelerate is not None:
    if self.pred_loss > loss and loss >= self.old_loss:
        self.current_lr = self.original_lr
        #print('restart')
    elif (loss-self.pred_loss).item() < 0.1:
        self.current_lr += .5*self.current_lr
        #print('powa')
    else:
        self.current_lr -= .5*self.current_lr
        #print(self.current_lr)
        #self.current_lr = min(loss.item()*3, 0.02)

    for g in self.optimizer.param_groups:
        g['lr'] = self.current_lr
'''



from abc import ABCMeta, abstractmethod


class AbstractProblem(metaclass=ABCMeta):

    @property
    def input_variables(self):
        variables = []

        if hasattr(self, 'spatial_variables'):
            variables += self.spatial_variables
        if hasattr(self, 'temporal_variable'):
            variables += self.temporal_variable
        if hasattr(self, 'parameters'):
            variables += self.parameters
        if hasattr(self, 'custom_variables'):
            variables += self.custom_variables

        return variables

    @input_variables.setter
    def input_variables(self, variables):
        raise NotImplementedError

    @property
    @abstractmethod
    def output_variables(self):
        pass

    @property
    @abstractmethod
    def conditions(self):
        pass



import torch

from pina.problem import SpatialProblem, ParametricProblem
from pina.operators import nabla
from pina import Span, Condition


class ParametricPoisson(SpatialProblem, ParametricProblem):

    spatial_variables = ['x', 'y']
    parameters = ['mu1', 'mu2']
    output_variables = ['u']
    domain = Span({'x': [-1, 1], 'y': [-1, 1]})

    def laplace_equation(input_, output_):
        force_term = torch.exp(
                - 2*(input_['x'] - input_['mu1'])**2 - 2*(input_['y'] -
                                                          input_['mu2'])**2)
        return nabla(output_['u'], input_) - force_term

    def nil_dirichlet(input_, output_):
        value = 0.0
        return output_['u'] - value

    conditions = {
        'gamma1': Condition(
            Span({'x': [-1, 1], 'y': 1, 'mu1': [-1, 1], 'mu2': [-1, 1]}),
            nil_dirichlet),
        'gamma2': Condition(
            Span({'x': [-1, 1], 'y': -1, 'mu1': [-1, 1], 'mu2': [-1, 1]}),
            nil_dirichlet),
        'gamma3': Condition(
            Span({'x': 1, 'y': [-1, 1], 'mu1': [-1, 1], 'mu2': [-1, 1]}),
            nil_dirichlet),
        'gamma4': Condition(
            Span({'x': -1, 'y': [-1, 1], 'mu1': [-1, 1], 'mu2': [-1, 1]}),
            nil_dirichlet),
        'D': Condition(
            Span({'x': [-1, 1], 'y': [-1, 1], 'mu1': [-1, 1], 'mu2': [-1, 1]}),
            laplace_equation),
    }



""" """
import torch

from .location import Location

class Condition:
    def __init__(self, *args, **kwargs):

        if len(args) == 2 and not kwargs:

            if (isinstance(args[0], torch.Tensor) and
                    isinstance(args[1], torch.Tensor)):
                self.input_points = args[0]
                self.output_points = args[1]
            elif isinstance(args[0], Location) and callable(args[1]):
                self.location = args[0]
                self.function = args[1]
            else:
                raise ValueError

        elif not args and len(kwargs) == 2:

            if 'input_points' in kwargs and 'output_points' in kwargs:
                self.input_points = kwargs['input_points']
                self.output_points = kwargs['output_points']
            elif 'location' in kwargs and 'function' in kwargs:
                self.location = kwargs['location']
                self.function = kwargs['function']
            else:
                raise ValueError
        else:
            raise ValueError



import torch

from .feed_forward import FeedForward


class MultiFeedForward(torch.nn.Module):

    def __init__(self, dff_dict):
        '''
        '''
        super().__init__()

        if not isinstance(dff_dict, dict):
            raise TypeError

        for name, constructor_args in dff_dict.items():
            setattr(self, name, FeedForward(**constructor_args))



import torch

class LabelTensor():

    def __init__(self, x, labels):


        if len(labels) != x.shape[1]:
            print(len(labels), x.shape[1])
            raise ValueError
        self.__labels = labels
        self.tensor = x

    def __getitem__(self, key):
        if key in self.labels:
            return self.tensor[:, self.labels.index(key)]
        else:
            return self.tensor.__getitem__(key)

    def __repr__(self):
        return self.tensor

    def __str__(self):
        return '{}\n {}\n'.format(self.labels, self.tensor)

    @property
    def shape(self):
        return self.tensor.shape

    @property
    def dtype(self):
        return self.tensor.dtype

    @property
    def device(self):
        return self.tensor.device

    @property
    def labels(self):
        return self.__labels

    @staticmethod
    def hstack(labeltensor_list):
        concatenated_tensor = torch.cat([lt.tensor for lt in labeltensor_list], axis=1)
        concatenated_label = sum([lt.labels for lt in labeltensor_list], [])
        return LabelTensor(concatenated_tensor, concatenated_label)



if __name__ == "__main__":
    import numpy as np
    a = np.random.uniform(size=(20, 3))
    a = np.random.uniform(size=(20, 3))
    p = torch.from_numpy(a)
    t = LabelTensor(p, labels=['u', 'p', 't'])
    print(t)
    print(t['u'])
    t *= 2
    print(t['u'])
    print(t[:, 0])




import numpy as np
def chebyshev_roots(n):
    """ Return the roots of *n* Chebyshev polynomials (between [-1, 1]) """
    coefficents = np.zeros(n+1)
    coefficents[-1] = 1
    return np.polynomial.chebyshev.chebroots(coefficents)




"""Module for operators vectorize implementation"""
import torch

from pina.label_tensor import LabelTensor


def grad(output_, input_):
    """
    TODO
    """
    if not isinstance(input_, LabelTensor):
        raise TypeError

    gradients = torch.autograd.grad(
        output_,
        input_.tensor,
        grad_outputs=torch.ones(output_.size()).to(
            dtype=input_.tensor.dtype,
            device=input_.tensor.device),
        create_graph=True, retain_graph=True, allow_unused=True)[0]
    return LabelTensor(gradients, input_.labels)


def div(output_, input_):
    """
    TODO
    """
    if output_.tensor.shape[1] == 1:
        div = grad(output_.tensor, input_).sum(axis=1)
    else:  # really to improve
        a = []
        for o in output_.tensor.T:
            a.append(grad(o, input_).tensor)
        div = torch.zeros(output_.tensor.shape[0], 1)
        for i in range(output_.tensor.shape[1]):
            div += a[i][:, i].reshape(-1, 1)

    return div


def nabla(output_, input_):
    """
    TODO
    """
    return div(grad(output_, input_), input_)



import numpy as np
import torch

from pina.problem import SpatialProblem
from pina.operators import nabla
from pina import Condition, Span


class Poisson(SpatialProblem):

    spatial_variables = ['x', 'y']
    output_variables = ['u']
    domain = Span({'x': [0, 1], 'y': [0, 1]})

    def laplace_equation(input_, output_):
        force_term = (torch.sin(input_['x']*torch.pi) *
                      torch.sin(input_['y']*torch.pi))
        return nabla(output_['u'], input_).flatten() - force_term

    def nil_dirichlet(input_, output_):
        value = 0.0
        return output_['u'] - value

    conditions = {
        'gamma1': Condition(Span({'x': [-1, 1], 'y':  1}), nil_dirichlet),
        'gamma2': Condition(Span({'x': [-1, 1], 'y': -1}), nil_dirichlet),
        'gamma3': Condition(Span({'x':  1, 'y': [-1, 1]}), nil_dirichlet),
        'gamma4': Condition(Span({'x': -1, 'y': [-1, 1]}), nil_dirichlet),
        'D': Condition(Span({'x': [-1, 1], 'y': [-1, 1]}), laplace_equation),
    }

    def poisson_sol(self, x, y):
        return -(np.sin(x*np.pi)*np.sin(y*np.pi))/(2*np.pi**2)

    truth_solution = poisson_sol



import numpy as np
from .chebyshev import chebyshev_roots
import torch

from .location import Location
from .label_tensor import LabelTensor


class Span(Location):
    def __init__(self, span_dict):

        self.fixed_ = {}
        self.range_ = {}

        for k, v in span_dict.items():
            if isinstance(v, (int, float)):
                self.fixed_[k] = v
            elif isinstance(v, (list, tuple)) and len(v) == 2:
                self.range_[k] = v
            else:
                raise TypeError

    def sample(self, n, mode='random'):

        bounds = np.array(list(self.range_.values()))
        if mode == 'random':
            pts = np.random.uniform(size=(n, bounds.shape[0]))
        elif mode == 'chebyshev':
            pts = np.array([
                chebyshev_roots(n) * .5 + .5
                for _ in range(bounds.shape[0])])
            grids = np.meshgrid(*pts)
            pts = np.hstack([grid.reshape(-1, 1) for grid in grids])
        elif mode == 'grid':
            pts = np.array([
                np.linspace(0, 1, n)
                for _ in range(bounds.shape[0])])
            grids = np.meshgrid(*pts)
            pts = np.hstack([grid.reshape(-1, 1) for grid in grids])
        elif mode == 'lh' or mode == 'latin':
            from scipy.stats import qmc
            sampler = qmc.LatinHypercube(d=bounds.shape[0])
            pts = sampler.random(n)

        # Scale pts
        pts *= bounds[:, 1] - bounds[:, 0]
        pts += bounds[:, 0]
        pts = torch.from_numpy(pts)
        pts_range_ = LabelTensor(pts, list(self.range_.keys()))

        fixed = torch.Tensor(list(self.fixed_.values()))
        pts_fixed_ = torch.ones(pts_range_.tensor.shape[0], len(self.fixed_)) * fixed
        pts_fixed_ = LabelTensor(pts_fixed_, list(self.fixed_.keys()))

        if self.fixed_:
            return LabelTensor.hstack([pts_range_, pts_fixed_])
        else:
            return pts_range_

    def meshgrid(self, n):
        pts = np.array([
            np.linspace(0, 1, n)
            for _ in range(self.bound.shape[0])])

        pts *= self.bound[:, 1] - self.bound[:, 0]
        pts += self.bound[:, 0]

        return np.meshgrid(*pts)



import numpy as np
import torch
import argparse
from pina.pinn import PINN
from pina.ppinn import ParametricPINN as pPINN
from pina.label_tensor import LabelTensor
from torch.nn import ReLU, Tanh, Softplus
from pina.adaptive_functions.adaptive_softplus import AdaptiveSoftplus
from problems.parametric_elliptic_optimal_control_alpha_variable import ParametricEllipticOptimalControl
from pina.multi_deep_feed_forward import MultiDeepFeedForward
from pina.deep_feed_forward import DeepFeedForward

alpha = 1

class myFeature(torch.nn.Module):
    """
    Feature: sin(x)
    """

    def __init__(self):
        super(myFeature, self).__init__()

    def forward(self, x):
        return (-x[:, 0]**2+1) * (-x[:, 1]**2+1)


class CustomMultiDFF(MultiDeepFeedForward):

    def __init__(self, dff_dict):
        super().__init__(dff_dict)

    def forward(self, x):
        out = self.uu(x)
        p = LabelTensor((out['u_param'] * x[:, 3]).reshape(-1, 1), ['p'])
        a = LabelTensor.hstack([out, p])
        return a

model = CustomMultiDFF(
        {
            'uu': {
                'input_variables': ['x1', 'x2', 'mu', 'alpha'],
                'output_variables': ['u_param', 'y'],
                'layers': [40, 40, 20],
                'func': Softplus,
                'extra_features': [myFeature()],
            },
            # 'u_param': {
            #     'input_variables': ['u', 'mu'],
            #     'output_variables': ['u_param'],
            #     'layers': [],
            #     'func': None
            # },
            # 'p': {
            # 'input_variables': ['u'],
            # 'output_variables': ['p'],
            # 'layers': [10],
            # 'func': None
            # },
        }
)


opc = ParametricEllipticOptimalControl(alpha)

if __name__ == "__main__":

    parser = argparse.ArgumentParser(description="Run PINA")
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument("-s", "-save", action="store_true")
    group.add_argument("-l", "-load", action="store_true")
    args = parser.parse_args()

    # model = DeepFeedForward(
    #     layers=[40, 40, 20],
    #     output_variables=['u_param', 'y', 'p'],
    #     input_variables=opc.input_variables+['mu', 'alpha'],
    #     func=Softplus,
    #     extra_features=[myFeature()]
    # )


    pinn = pPINN(
        opc,
        model,
        lr=0.002,
        error_norm='mse',
        regularizer=1e-8,
        lr_accelerate=None)

    if args.s:

        pinn.span_pts(30, 'grid', ['D1'])
        pinn.span_pts(50, 'grid', ['gamma1', 'gamma2', 'gamma3', 'gamma4'])
        pinn.train(10000, 20)
        # with open('ocp_wrong_history.txt', 'w') as file_:
        #     for i, losses in enumerate(pinn.history):
        #         file_.write('{} {}\n'.format(i, sum(losses).item()))

        pinn.save_state('pina.ocp')

    else:
        pinn.load_state('working.pina.ocp')
        pinn.load_state('pina.ocp')

        import matplotlib
        matplotlib.use('GTK3Agg')
        import matplotlib.pyplot as plt

        # res = 64
        # param = torch.tensor([[3., 1]])
        # pts_container = []
        # for mn, mx in [[-1, 1], [-1, 1]]:
        #     pts_container.append(np.linspace(mn, mx, res))
        # grids_container = np.meshgrid(*pts_container)
        # unrolled_pts = torch.tensor([t.flatten() for t in grids_container]).T
        # unrolled_pts = torch.cat([unrolled_pts, param.double().repeat(unrolled_pts.shape[0], 1).reshape(-1, 2)], axis=1)

        # unrolled_pts = LabelTensor(unrolled_pts, ['x1', 'x2', 'mu', 'alpha'])
        # Z_pred = pinn.model(unrolled_pts.tensor)
        # print(Z_pred.tensor.shape)

        # plt.subplot(2, 3, 1)
        # plt.pcolor(Z_pred['y'].reshape(res, res).detach())
        # plt.colorbar()
        # plt.subplot(2, 3, 2)
        # plt.pcolor(Z_pred['u_param'].reshape(res, res).detach())
        # plt.colorbar()
        # plt.subplot(2, 3, 3)
        # plt.pcolor(Z_pred['p'].reshape(res, res).detach())
        # plt.colorbar()
        # with open('ocp_mu3_a1_plot.txt', 'w') as f_:
        #     f_.write('x y u p ys\n')
        #     for (x, y), tru, pre, e in zip(unrolled_pts[:, :2],
        #                                    Z_pred['u_param'].reshape(-1, 1),
        #                                    Z_pred['p'].reshape(-1, 1),
        #                                    Z_pred['y'].reshape(-1, 1),
        #                                    ):
        #         f_.write('{} {} {} {} {}\n'.format(x.item(), y.item(), tru.item(), pre.item(), e.item()))


        # param = torch.tensor([[3.0, 0.01]])
        # unrolled_pts = torch.tensor([t.flatten() for t in grids_container]).T
        # unrolled_pts = torch.cat([unrolled_pts, param.double().repeat(unrolled_pts.shape[0], 1).reshape(-1, 2)], axis=1)
        # unrolled_pts = LabelTensor(unrolled_pts, ['x1', 'x2', 'mu', 'alpha'])
        # Z_pred = pinn.model(unrolled_pts.tensor)

        # plt.subplot(2, 3, 4)
        # plt.pcolor(Z_pred['y'].reshape(res, res).detach())
        # plt.colorbar()
        # plt.subplot(2, 3, 5)
        # plt.pcolor(Z_pred['u_param'].reshape(res, res).detach())
        # plt.colorbar()
        # plt.subplot(2, 3, 6)
        # plt.pcolor(Z_pred['p'].reshape(res, res).detach())
        # plt.colorbar()

        # plt.show()
        # with open('ocp_mu3_a0.01_plot.txt', 'w') as f_:
        #     f_.write('x y u p ys\n')
        #     for (x, y), tru, pre, e in zip(unrolled_pts[:, :2],
        #                                    Z_pred['u_param'].reshape(-1, 1),
        #                                    Z_pred['p'].reshape(-1, 1),
        #                                    Z_pred['y'].reshape(-1, 1),
        #                                    ):
        #         f_.write('{} {} {} {} {}\n'.format(x.item(), y.item(), tru.item(), pre.item(), e.item()))




        y = {}
        u = {}
        for alpha in [0.01, 0.1, 1]:
            y[alpha] = []
            u[alpha] = []
            for p in np.linspace(0.5, 3, 32):
                a = pinn.model(LabelTensor(torch.tensor([[0, 0, p, alpha]]).double(), ['x1', 'x2', 'mu', 'alpha']).tensor)
                y[alpha].append(a['y'].detach().numpy()[0])
                u[alpha].append(a['u_param'].detach().numpy()[0])



        plt.plot(np.linspace(0.5, 3, 32), u[1], label='u')
        plt.plot(np.linspace(0.5, 3, 32), u[0.01], label='u')
        plt.plot(np.linspace(0.5, 3, 32), u[0.1], label='u')
        plt.plot([1, 2, 3], [0.28, 0.56, 0.85], 'o', label='Truth values')
        plt.legend()
        plt.show()
        print(y[1])
        print(y[0.1])
        print(y[0.01])
        with open('elliptic_param_y.txt', 'w') as f_:
            f_.write('mu 1 01 001\n')
            for mu, y1, y01, y001 in zip(np.linspace(0.5, 3, 32), y[1], y[0.1], y[0.01]):
                f_.write('{} {} {} {}\n'.format(mu, y1, y01, y001))

        with open('elliptic_param_u.txt', 'w') as f_:
            f_.write('mu 1 01 001\n')
            for mu, y1, y01, y001 in zip(np.linspace(0.5, 3, 32), u[1], u[0.1], u[0.01]):
                f_.write('{} {} {} {}\n'.format(mu, y1, y01, y001))


        plt.plot(np.linspace(0.5, 3, 32), y, label='y')
        plt.plot([1, 2, 3], [0.062, 0.12, 0.19], 'o', label='Truth values')
        plt.legend()
        plt.show()





import argparse
import sys
import numpy as np
import torch
from torch.nn import ReLU, Tanh, Softplus

from pina import PINN, LabelTensor, Plotter
from pina.model import FeedForward
from pina.adaptive_functions import AdaptiveSin, AdaptiveCos, AdaptiveTanh
from problems.poisson import Poisson


class myFeature(torch.nn.Module):
    """
    Feature: sin(x)
    """

    def __init__(self):
        super(myFeature, self).__init__()

    def forward(self, x):
        return torch.sin(x[:, 0]*torch.pi) * torch.sin(x[:, 1]*torch.pi)

if __name__ == "__main__":

    parser = argparse.ArgumentParser(description="Run PINA")
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument("-s", "-save", action="store_true")
    group.add_argument("-l", "-load", action="store_true")
    parser.add_argument("id_run", help="number of run", type=int)
    parser.add_argument("features", help="extra features", type=int)
    args = parser.parse_args()

    feat = [myFeature()] if args.features else []

    poisson_problem = Poisson()
    model = FeedForward(
        layers=[10, 10],
        output_variables=poisson_problem.output_variables,
        input_variables=poisson_problem.input_variables,
        func=Softplus,
        extra_features=feat
    )

    pinn = PINN(
        poisson_problem,
        model,
        lr=0.003,
        error_norm='mse',
        regularizer=1e-8,
        lr_accelerate=None)

    if args.s:

        pinn.span_pts(20, 'grid', ['D'])
        pinn.span_pts(20, 'grid', ['gamma1', 'gamma2', 'gamma3', 'gamma4'])
        #pinn.plot_pts()
        pinn.train(1000, 100)
        with open('poisson_history_{}_{}.txt'.format(args.id_run, args.features), 'w') as file_:
            for i, losses in enumerate(pinn.history):
                file_.write('{} {}\n'.format(i, sum(losses)))
        pinn.save_state('pina.poisson')

    else:
        pinn.load_state('pina.poisson')
        plotter = Plotter()
        plotter.plot(pinn)





import numpy as np
import torch
from pina.problem import Problem
from pina.segment import Segment
from pina.cube import Cube
from pina.problem2d import Problem2D

xmin, xmax, ymin, ymax = -1, 1, -1, 1

class EllipticOptimalControl(Problem2D):

    def __init__(self, alpha=1):

        def term1(input_, output_):
            grad_p = self.grad(output_['p'], input_)
            gradgrad_p_x1 = self.grad(grad_p['x1'], input_)
            gradgrad_p_x2 = self.grad(grad_p['x2'], input_)
            yd = 2.0
            return output_['y'] - yd - (gradgrad_p_x1['x1'] + gradgrad_p_x2['x2'])

        def term2(input_, output_):
            grad_y = self.grad(output_['y'], input_)
            gradgrad_y_x1 = self.grad(grad_y['x1'], input_)
            gradgrad_y_x2 = self.grad(grad_y['x2'], input_)
            return - (gradgrad_y_x1['x1'] + gradgrad_y_x2['x2']) - output_['u']

        def term3(input_, output_):
            return output_['p'] - output_['u']*alpha


        def nil_dirichlet(input_, output_):
            y_value = 0.0
            p_value = 0.0
            return torch.abs(output_['y'] - y_value) + torch.abs(output_['p'] - p_value)

        self.conditions = {
            'gamma1': {'location': Segment((xmin, ymin), (xmax, ymin)), 'func': nil_dirichlet},
            'gamma2': {'location': Segment((xmax, ymin), (xmax, ymax)), 'func': nil_dirichlet},
            'gamma3': {'location': Segment((xmax, ymax), (xmin, ymax)), 'func': nil_dirichlet},
            'gamma4': {'location': Segment((xmin, ymax), (xmin, ymin)), 'func': nil_dirichlet},
            'D1': {'location': Cube([[xmin, xmax], [ymin, ymax]]), 'func': [term1, term2, term3]},
            #'D2': {'location': Cube([[0, 1], [0, 1]]), 'func': term2},
            #'D3': {'location': Cube([[0, 1], [0, 1]]), 'func': term3}
        }

        self.input_variables = ['x1', 'x2']
        self.output_variables = ['u', 'p', 'y']
        self.spatial_domain = Cube([[xmin, xmax], [xmin, xmax]])




""" Module for plotting. """
import matplotlib
#matplotlib.use('Qt5Agg')
import matplotlib.pyplot as plt
import numpy as np
import torch

from pina import LabelTensor
from pina import PINN
from .problem import SpatialProblem, TimeDependentProblem
#from pina.tdproblem1d import TimeDepProblem1D


class Plotter:

    def _plot_2D(self, obj, method='contourf'):
        """
        """
        if not isinstance(obj, PINN):
            raise RuntimeError

        res = 256
        pts = obj.problem.spatial_domain.discretize(res, 'grid')
        grids_container = [
            pts[:, 0].reshape(res, res),
            pts[:, 1].reshape(res, res),
        ]
        pts = LabelTensor(torch.tensor(pts), obj.problem.input_variables)
        predicted_output = obj.model(pts.tensor)

        if hasattr(obj.problem, 'truth_solution'):
            truth_output = obj.problem.truth_solution(*pts.tensor.T).float()
            fig, axes = plt.subplots(nrows=1, ncols=3, figsize=(16, 6))

            cb = getattr(axes[0], method)(*grids_container, predicted_output.tensor.reshape(res, res).detach())
            fig.colorbar(cb, ax=axes[0])
            cb = getattr(axes[1], method)(*grids_container, truth_output.reshape(res, res).detach())
            fig.colorbar(cb, ax=axes[1])
            cb = getattr(axes[2], method)(*grids_container, (truth_output-predicted_output.tensor.float().flatten()).detach().reshape(res, res))
            fig.colorbar(cb, ax=axes[2])
        else:
            fig, axes = plt.subplots(nrows=1, ncols=1, figsize=(8, 6))
            cb = getattr(axes, method)(*grids_container, predicted_output.tensor.reshape(res, res).detach())
            fig.colorbar(cb, ax=axes)


    def _plot_1D_TimeDep(self, obj, method='contourf'):
        """
        """
        if not isinstance(obj, PINN):
            raise RuntimeError

        res = 256
        grids_container = np.meshgrid(
            obj.problem.spatial_domain.discretize(res, 'grid'),
            obj.problem.temporal_domain.discretize(res, 'grid'),
        )
        pts = np.hstack([
            grids_container[0].reshape(-1, 1),
            grids_container[1].reshape(-1, 1),
        ])
        pts = LabelTensor(torch.tensor(pts), obj.problem.input_variables)
        predicted_output = obj.model(pts.tensor)

        if hasattr(obj.problem, 'truth_solution'):
            truth_output = obj.problem.truth_solution(*pts.tensor.T).float()
            fig, axes = plt.subplots(nrows=1, ncols=3, figsize=(16, 6))

            cb = getattr(axes[0], method)(*grids_container, predicted_output.tensor.reshape(res, res).detach())
            fig.colorbar(cb, ax=axes[0])
            cb = getattr(axes[1], method)(*grids_container, truth_output.reshape(res, res).detach())
            fig.colorbar(cb, ax=axes[1])
            cb = getattr(axes[2], method)(*grids_container, (truth_output-predicted_output.tensor.float().flatten()).detach().reshape(res, res))
            fig.colorbar(cb, ax=axes[2])
        else:
            fig, axes = plt.subplots(nrows=1, ncols=1, figsize=(8, 6))
            cb = getattr(axes, method)(*grids_container, predicted_output.tensor.reshape(res, res).detach())
            fig.colorbar(cb, ax=axes)



    def plot(self, obj, method='contourf', filename=None):
        """
        """
        res = 256
        pts = obj.problem.domain.sample(res, 'grid')
        grids_container = [
            pts[:, 0].reshape(res, res),
            pts[:, 1].reshape(res, res),
        ]
        predicted_output = obj.model(pts)

        if hasattr(obj.problem, 'truth_solution'):
            truth_output = obj.problem.truth_solution(*pts.tensor.T).float()
            fig, axes = plt.subplots(nrows=1, ncols=3, figsize=(16, 6))

            cb = getattr(axes[0], method)(*grids_container, predicted_output.tensor.reshape(res, res).detach())
            fig.colorbar(cb, ax=axes[0])
            cb = getattr(axes[1], method)(*grids_container, truth_output.reshape(res, res).detach())
            fig.colorbar(cb, ax=axes[1])
            cb = getattr(axes[2], method)(*grids_container, (truth_output-predicted_output.tensor.float().flatten()).detach().reshape(res, res))
            fig.colorbar(cb, ax=axes[2])
        else:
            fig, axes = plt.subplots(nrows=1, ncols=1, figsize=(8, 6))
            cb = getattr(axes, method)(*grids_container, predicted_output.tensor.reshape(res, res).detach())
            fig.colorbar(cb, ax=axes)

        if filename:
            plt.savefig(filename)
        else:
            plt.show()



__all__ = [
    'PINN',
    'ParametricPINN',
    'LabelTensor',
    'Plotter',
    'Condition',
    'Span'
]

from .label_tensor import LabelTensor
from .pinn import PINN
#from .ppinn import ParametricPINN
from .plotter import Plotter
from .span import Span
from .condition import Condition



from .problem import AbstractProblem
import torch
import matplotlib.pyplot as plt
import numpy as np
from pina.label_tensor import LabelTensor
torch.pi = torch.acos(torch.zeros(1)).item() * 2 # which is 3.1415927410125732

class PINN(object):

    def __init__(self,
            problem,
            model,
            optimizer=torch.optim.Adam,
            lr=0.001,
            regularizer=0.00001,
            data_weight=1.,
            dtype=torch.float64,
            device='cpu',
            lr_accelerate=None,
            error_norm='mse'):
        '''
        :param Problem problem: the formualation of the problem.
        :param dict architecture: a dictionary containing the information to
            build the model. Valid options are:
            - inner_size [int] the number of neurons in the hidden layers; by
              default is 20.
            - n_layers [int] the number of hidden layers; by default is 4.
            - func [nn.Module or str] the activation function; passing a `str`
              is possible to chose adaptive function (between 'adapt_tanh'); by
              default is non-adaptive iperbolic tangent.
        :param float lr: the learning rate; default is 0.001
        :param float regularizer: the coefficient for L2 regularizer term
        :param type dtype: the data type to use for the model. Valid option are
            `torch.float32` and `torch.float64` (`torch.float16` only on GPU);
            default is `torch.float64`.
        :param float lr_accelete: the coefficient that controls the learning
            rate increase, such that, for all the epoches in which the loss is
            decreasing, the learning_rate is update using
                $learning_rate = learning_rate * lr_accelerate$.
            When the loss stops to decrease, the learning rate is set to the
            initial value [TODO test parameters]

        '''

        self.problem = problem


        # self._architecture = architecture if architecture else dict()
        # self._architecture['input_dimension'] = self.problem.domain_bound.shape[0]
        # self._architecture['output_dimension'] = len(self.problem.variables)
        # if hasattr(self.problem, 'params_domain'):
            # self._architecture['input_dimension'] += self.problem.params_domain.shape[0]

        self.accelerate = lr_accelerate

        self.error_norm = error_norm

        if device == 'cuda' and not torch.cuda.is_available():
            raise RuntimeError
        self.device = torch.device(device)

        self.dtype = dtype
        self.history = []

        self.model = model
        self.model.to(dtype=self.dtype, device=self.device)

        self.input_pts = {}
        self.truth_values = {}


        self.trained_epoch = 0
        self.optimizer = optimizer(
            self.model.parameters(), lr=lr, weight_decay=regularizer)

        self.data_weight = data_weight

    @property
    def problem(self):
        return self._problem

    @problem.setter
    def problem(self, problem):
        if not isinstance(problem, AbstractProblem):
            raise TypeError
        self._problem = problem

    def get_data_residuals(self):

        data_residuals = []

        for output in self.data_pts:
            data_values_pred = self.model(self.data_pts[output])
            data_residuals.append(data_values_pred - self.data_values[output])

        return torch.cat(data_residuals)

    def get_phys_residuals(self):
        """
        """

        residuals = []
        for equation in self.problem.equation:
            residuals.append(equation(self.phys_pts, self.model(self.phys_pts)))
        return residuals


    def _compute_norm(self, vec):
        """
        Compute the norm of the `vec` one-dimensional tensor based on the
        `self.error_norm` attribute.

        .. todo: complete

        :param vec torch.tensor: the tensor
        """
        if isinstance(self.error_norm, int):
            return torch.sum(torch.abs(vec**self.error_norm))**(1./self.error_norm)
        elif self.error_norm == 'mse':
            return torch.mean(vec**2)
        elif self.error_norm == 'me':
            return torch.mean(torch.abs(vec))
        else:
            raise RuntimeError

    def save_state(self, filename):

        checkpoint = {
                'epoch': self.trained_epoch,
                'model_state': self.model.state_dict(),
                'optimizer_state' : self.optimizer.state_dict(),
                'optimizer_class' : self.optimizer.__class__,
                'history' : self.history,
        }

        # TODO save also architecture param?
        #if isinstance(self.model, DeepFeedForward):
        #    checkpoint['model_class'] = self.model.__class__
        #    checkpoint['model_structure'] = {
        #    }
        torch.save(checkpoint, filename)

    def load_state(self, filename):

        checkpoint = torch.load(filename)
        self.model.load_state_dict(checkpoint['model_state'])


        self.optimizer = checkpoint['optimizer_class'](self.model.parameters())
        self.optimizer.load_state_dict(checkpoint['optimizer_state'])

        self.trained_epoch = checkpoint['epoch']
        self.history = checkpoint['history']

        return self


    def span_pts(self, n, mode='grid', locations='all'):
        '''

        '''

        if locations == 'all':
            locations = [condition for condition in self.problem.conditions]

        for location in locations:
            condition = self.problem.conditions[location]

            try:
                pts = condition.location.sample(n, mode)
            except:
                pts = condition.input_points

            self.input_pts[location] = pts
            print(pts.tensor.shape)
            self.input_pts[location].tensor.to(dtype=self.dtype, device=self.device)
            self.input_pts[location].tensor.requires_grad_(True)
            self.input_pts[location].tensor.retain_grad()


    def plot_pts(self, locations='all'):
        import matplotlib
        matplotlib.use('GTK3Agg')
        if locations == 'all':
            locations = [condition for condition in self.problem.conditions]

        for location in locations:
            x, y = self.input_pts[location].tensor.T
            #plt.plot(x.detach(), y.detach(), 'o', label=location)
            np.savetxt('burgers_{}_pts.txt'.format(location), self.input_pts[location].tensor.detach(), header='x y', delimiter=' ')


        plt.legend()
        plt.show()



    def train(self, stop=100, frequency_print=2, trial=None):

        epoch = 0

        while True:

            losses = []

            for condition_name in self.problem.conditions:
                condition = self.problem.conditions[condition_name]
                pts = self.input_pts[condition_name]

                predicted = self.model(pts)

                residuals = condition.function(pts, predicted)
                losses.append(self._compute_norm(residuals))

            self.optimizer.zero_grad()
            sum(losses).backward()
            self.optimizer.step()

            self.trained_epoch += 1
            if epoch % 50 == 0:
                self.history.append([loss.detach().item() for loss in losses])
            epoch += 1

            if trial:
                import optuna
                trial.report(sum(losses), epoch)
                if trial.should_prune():
                    raise optuna.exceptions.TrialPruned()

            if isinstance(stop, int):
                if epoch == stop:
                    break
            elif isinstance(stop, float):
                if sum(losses) < stop:
                    break

            if epoch % frequency_print == 0:
                print('[epoch {:05d}] {:.6e} '.format(self.trained_epoch, sum(losses).item()), end='')
                for loss in losses:
                    print('{:.6e} '.format(loss), end='')
                print()

        return sum(losses).item()


    def error(self, dtype='l2', res=100):

        import numpy as np
        if hasattr(self.problem, 'truth_solution') and self.problem.truth_solution is not None:
            pts_container = []
            for mn, mx in self.problem.domain_bound:
                pts_container.append(np.linspace(mn, mx, res))
            grids_container = np.meshgrid(*pts_container)
            Z_true = self.problem.truth_solution(*grids_container)

        elif hasattr(self.problem, 'data_solution') and self.problem.data_solution is not None:
            grids_container = self.problem.data_solution['grid']
            Z_true = self.problem.data_solution['grid_solution']
        try:
            unrolled_pts = torch.tensor([t.flatten() for t in grids_container]).T.to(dtype=self.dtype, device=self.device)
            Z_pred = self.model(unrolled_pts)
            Z_pred = Z_pred.detach().numpy().reshape(grids_container[0].shape)

            if dtype == 'l2':
                return np.linalg.norm(Z_pred - Z_true)/np.linalg.norm(Z_true)
            else:
                # TODO H1
                pass
        except:
            print("")
            print("Something went wrong...")
            print("Not able to compute the error. Please pass a data solution or a true solution")




    def plot(self, res, filename=None, variable=None):
            '''
            '''
            import matplotlib
            matplotlib.use('Agg')
            import matplotlib.pyplot as plt

            self._plot_2D(res, filename, variable)
            print('TTTTTTTTTTTTTTTTTt')
            print(self.problem.bounds)
            pts_container = []
            #for mn, mx in [[-1, 1], [-1, 1]]:
            for mn, mx in [[0, 1], [0, 1]]:
            #for mn, mx in [[-1, 1], [0, 1]]:
                pts_container.append(np.linspace(mn, mx, res))
            grids_container = np.meshgrid(*pts_container)
            unrolled_pts = torch.tensor([t.flatten() for t in grids_container]).T
            unrolled_pts.to(dtype=self.dtype)
            Z_pred = self.model(unrolled_pts)

            #######################################################
            # poisson
            # Z_truth = self.problem.truth_solution(unrolled_pts[:, 0], unrolled_pts[:, 1])
            # Z_pred = Z_pred.tensor.detach().reshape(grids_container[0].shape)
            # Z_truth = Z_truth.detach().reshape(grids_container[0].shape)
            # err = np.abs(Z_pred-Z_truth)


            # with open('poisson2_nofeat_plot.txt', 'w') as f_:
            #     f_.write('x y truth pred e\n')
            #     for (x, y), tru, pre, e in zip(unrolled_pts, Z_truth.reshape(-1, 1), Z_pred.reshape(-1, 1), err.reshape(-1, 1)):
            #         f_.write('{} {} {} {} {}\n'.format(x.item(), y.item(), tru.item(), pre.item(), e.item()))
            # n = Z_pred.shape[1]
            # plt.figure(figsize=(16, 6))
            # plt.subplot(1, 3, 1)
            # plt.contourf(*grids_container, Z_truth)
            # plt.colorbar()
            # plt.subplot(1, 3, 2)
            # plt.contourf(*grids_container, Z_pred)
            # plt.colorbar()
            # plt.subplot(1, 3, 3)
            # plt.contourf(*grids_container, err)
            # plt.colorbar()
            # plt.show()

            #######################################################
            # burgers
            import scipy
            data = scipy.io.loadmat('Data/burgers_shock.mat')
            data_solution = {'grid': np.meshgrid(data['x'], data['t']), 'grid_solution': data['usol'].T}

            grids_container = data_solution['grid']
            print(data_solution['grid_solution'].shape)
            unrolled_pts = torch.tensor([t.flatten() for t in grids_container]).T
            unrolled_pts.to(dtype=self.dtype)
            Z_pred = self.model(unrolled_pts)
            Z_truth = data_solution['grid_solution']

            Z_pred = Z_pred.tensor.detach().reshape(grids_container[0].shape)
            print(Z_pred, Z_truth)
            err = np.abs(Z_pred.numpy()-Z_truth)


            with open('burgers_nofeat_plot.txt', 'w') as f_:
                f_.write('x y truth pred e\n')
                for (x, y), tru, pre, e in zip(unrolled_pts, Z_truth.reshape(-1, 1), Z_pred.reshape(-1, 1), err.reshape(-1, 1)):
                    f_.write('{} {} {} {} {}\n'.format(x.item(), y.item(), tru.item(), pre.item(), e.item()))
            n = Z_pred.shape[1]
            plt.figure(figsize=(16, 6))
            plt.subplot(1, 3, 1)
            plt.contourf(*grids_container, Z_truth,vmin=-1, vmax=1)
            plt.colorbar()
            plt.subplot(1, 3, 2)
            plt.contourf(*grids_container, Z_pred, vmin=-1, vmax=1)
            plt.colorbar()
            plt.subplot(1, 3, 3)
            plt.contourf(*grids_container, err)
            plt.colorbar()
            plt.show()


            # for i, output in enumerate(Z_pred.tensor.T, start=1):
            #     output = output.detach().numpy().reshape(grids_container[0].shape)
            #     plt.subplot(1, n, i)
            #     plt.contourf(*grids_container, output)
            #     plt.colorbar()

            if filename is None:
                plt.show()
            else:
                plt.savefig(filename)

    def plot_params(self, res, param, filename=None, variable=None):
            '''
            '''
            import matplotlib
            matplotlib.use('GTK3Agg')
            import matplotlib.pyplot as plt

            if hasattr(self.problem, 'truth_solution') and self.problem.truth_solution is not None:
                n_plot = 2
            elif hasattr(self.problem, 'data_solution') and self.problem.data_solution is not None:
                n_plot = 2
            else:
                n_plot = 1

            fig, axs = plt.subplots(nrows=1, ncols=n_plot, figsize=(n_plot*6,4))
            if not isinstance(axs, np.ndarray): axs = [axs]

            if hasattr(self.problem, 'data_solution') and self.problem.data_solution is not None:
                grids_container = self.problem.data_solution['grid']
                Z_true = self.problem.data_solution['grid_solution']
            elif hasattr(self.problem, 'truth_solution') and self.problem.truth_solution is not None:

                pts_container = []
                for mn, mx in self.problem.domain_bound:
                    pts_container.append(np.linspace(mn, mx, res))

                grids_container = np.meshgrid(*pts_container)
                Z_true = self.problem.truth_solution(*grids_container)

            pts_container = []
            for mn, mx in self.problem.domain_bound:
                pts_container.append(np.linspace(mn, mx, res))
            grids_container = np.meshgrid(*pts_container)
            unrolled_pts = torch.tensor([t.flatten() for t in grids_container]).T.to(dtype=self.type)
            #print(unrolled_pts)
            #print(param)
            param_unrolled_pts = torch.cat((unrolled_pts, param.repeat(unrolled_pts.shape[0], 1)), 1)
            if variable==None:
                variable = self.problem.variables[0]
                Z_pred = self.evaluate(param_unrolled_pts)[variable]
                variable = "Solution"
            else:
                Z_pred = self.evaluate(param_unrolled_pts)[variable]

            Z_pred= Z_pred.detach().numpy().reshape(grids_container[0].shape)
            set_pred = axs[0].contourf(*grids_container, Z_pred)
            axs[0].set_title('PINN [trained epoch = {}]'.format(self.trained_epoch) + " " + variable) #TODO add info about parameter in the title
            fig.colorbar(set_pred, ax=axs[0])

            if n_plot == 2:

                set_true = axs[1].contourf(*grids_container, Z_true)

                axs[1].set_title('Truth solution')
                fig.colorbar(set_true, ax=axs[1])

            if filename is None:
                    plt.show()
            else:
                    fig.savefig(filename + " " + variable)

    def plot_error(self, res, filename=None):
        import matplotlib
        matplotlib.use('GTK3Agg')
        import matplotlib.pyplot as plt


        fig, axs = plt.subplots(nrows=1, ncols=1, figsize=(6,4))
        if not isinstance(axs, np.ndarray): axs = [axs]

        if hasattr(self.problem, 'data_solution') and self.problem.data_solution is not None:
            grids_container = self.problem.data_solution['grid']
            Z_true = self.problem.data_solution['grid_solution']
        elif hasattr(self.problem, 'truth_solution') and self.problem.truth_solution is not None:
            pts_container = []
            for mn, mx in self.problem.domain_bound:
                pts_container.append(np.linspace(mn, mx, res))

            grids_container = np.meshgrid(*pts_container)
            Z_true = self.problem.truth_solution(*grids_container)
        try:
            unrolled_pts = torch.tensor([t.flatten() for t in grids_container]).T.to(dtype=self.type)

            Z_pred = self.model(unrolled_pts)
            Z_pred = Z_pred.detach().numpy().reshape(grids_container[0].shape)
            set_pred = axs[0].contourf(*grids_container, abs(Z_pred - Z_true))
            axs[0].set_title('PINN [trained epoch = {}]'.format(self.trained_epoch) + "Pointwise Error")
            fig.colorbar(set_pred, ax=axs[0])

            if filename is None:
                    plt.show()
            else:
                    fig.savefig(filename)
        except:
            print("")
            print("Something went wrong...")
            print("Not able to plot the error. Please pass a data solution or a true solution")

'''
print(self.pred_loss.item(),loss.item(), self.old_loss.item())
if self.accelerate is not None:
    if self.pred_loss > loss and loss >= self.old_loss:
        self.current_lr = self.original_lr
        #print('restart')
    elif (loss-self.pred_loss).item() < 0.1:
        self.current_lr += .5*self.current_lr
        #print('powa')
    else:
        self.current_lr -= .5*self.current_lr
        #print(self.current_lr)
        #self.current_lr = min(loss.item()*3, 0.02)

    for g in self.optimizer.param_groups:
        g['lr'] = self.current_lr
'''



__all__ = [
    '__project__',
    '__title__',
    '__author__',
    '__copyright__',
    '__license__',
    '__version__',
    '__mail__',
    '__maintainer__',
    '__status__']

__project__ = 'PINA'
__title__ = "pina"
__author__ = "Nicola Demo, Maria Strazzullo"
__copyright__ = "Copyright 2021-2021, PINA contributors"
__license__ = "MIT"
__version__ = "0.0.1"
__mail__ = 'demo.nicola@gmail.com, '  # TODO
__maintainer__ = __author__
__status__ = "Alpha"
__packagename__ = "pina-mathlab"




import argparse
import torch
from torch.nn import Softplus

from pina import PINN as pPINN
from problems.parametric_poisson import ParametricPoisson
from pina.model import FeedForward


class myFeature(torch.nn.Module):
    """
    """
    def __init__(self):
        super(myFeature, self).__init__()

    def forward(self, x):
        return torch.exp(- 2*(x['x'] - x['mu1'])**2 - 2*(x['y'] - x['mu2'])**2)


if __name__ == "__main__":

    parser = argparse.ArgumentParser(description="Run PINA")
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument("-s", "-save", action="store_true")
    group.add_argument("-l", "-load", action="store_true")
    parser.add_argument("id_run", help="number of run", type=int)
    parser.add_argument("features", help="extra features", type=int)
    args = parser.parse_args()

    feat = [myFeature()] if args.features else []

    poisson_problem = ParametricPoisson()
    model = FeedForward(
        layers=[200, 40, 10],
        output_variables=poisson_problem.output_variables,
        input_variables=poisson_problem.input_variables,
        func=Softplus,
        extra_features=feat
    )

    pinn = pPINN(
        poisson_problem,
        model,
        lr=0.0006,
        regularizer=1e-6,
        lr_accelerate=None)

    if args.s:

        pinn.span_pts(2000, 'random', ['D'])
        pinn.span_pts(200, 'random', ['gamma1', 'gamma2', 'gamma3', 'gamma4'])
        pinn.train(10000, 10)
        pinn.save_state('pina.poisson_param')

    else:
        pinn.load_state('pina.poisson_param')



import torch
import torch.nn as nn

from pina.label_tensor import LabelTensor


class FeedForward(torch.nn.Module):

    def __init__(self, input_variables, output_variables, inner_size=20,
                 n_layers=2, func=nn.Tanh, layers=None, extra_features=None):
        '''
        '''
        super().__init__()

        if extra_features is None:
            extra_features = []
        self.extra_features = nn.Sequential(*extra_features)

        self.input_variables = input_variables
        self.input_dimension = len(input_variables)

        self.output_variables = output_variables
        self.output_dimension = len(output_variables)

        n_features = len(extra_features)

        if layers is None:
            layers = [inner_size] * n_layers

        tmp_layers = layers.copy()
        tmp_layers.insert(0, self.input_dimension+n_features)
        tmp_layers.append(self.output_dimension)

        self.layers = []
        for i in range(len(tmp_layers)-1):
            self.layers.append(nn.Linear(tmp_layers[i], tmp_layers[i+1]))

        if isinstance(func, list):
            self.functions = func
        else:
            self.functions = [func for _ in range(len(self.layers)-1)]

        unique_list = []
        for layer, func in zip(self.layers[:-1], self.functions):
            unique_list.append(layer)
            if func is not None:
                unique_list.append(func())
        unique_list.append(self.layers[-1])

        self.model = nn.Sequential(*unique_list)

    def forward(self, x):
        """
        """
        nf = len(self.extra_features)
        if nf == 0:
            return LabelTensor(self.model(x.tensor), self.output_variables)

        # if self.extra_features
        input_ = torch.zeros(x.shape[0], nf+x.shape[1], dtype=x.dtype,
                             device=x.device)
        input_[:, :x.shape[1]] = x.tensor
        for i, feature in enumerate(self.extra_features,
                                    start=self.input_dimension):
            input_[:, i] = feature(x)
        return LabelTensor(self.model(input_), self.output_variables)
